/******************************************************************************/
/*                                                                            */
/* Version 1.0                                                                */
/* Author: Thomas Sallantin                Date: April 2001                   */
/*                                                                            */
/******************************************************************************/
/* Version 1.0   First release NT CheckDuplication                           */
/******************************************************************************/
/*                                                                            */
/* Note: This plugin node implementation is not necessarily portable          */
/* to all platforms supported by MQSeries Integrator Version 2.               */
/*                                                                            */
/******************************************************************************/


/* Includes */
#include <stdio.h>
#include <windows.h>
#include <string.h>

#include "BipSampPluginUtil.h"
#include "CheckDuplication.h"

FILE *outputFile = NULL;

/******************************************************************************/
/*                                                                            */
/* Plugin Node Utility function:                                              */
/*                                                                            */
/*  Acquire storage for attribute and link to the context                     */
/*                                                                            */
/******************************************************************************/
ATTRIBUTE_TBL_ENTRY *insertAttributeTableEntry(NODE_CONTEXT_ST *context, wchar_t *nodeName, CciChar *attrName, int type)
{
   static wchar_t* functionName=L"insertAttributeTableEntry()";
   ATTRIBUTE_TBL_ENTRY* entry;
   int                  attrSize;

   /* Allocate storage for the attribute entry */
   entry = malloc(sizeof(ATTRIBUTE_TBL_ENTRY));

   if (entry) {
      entry->next = 0;
      entry->type = type;

      switch (entry->type) {
        case CNI_TYPE_INTEGER:
            attrSize = 4;
            break;
        case CNI_TYPE_STRING:
            attrSize = MAX_STRING_SIZE;
        default:
            /* Error: Invalid datatype */
            break;
      } /* switch (entry->type) */

      /* Acquire and clear storage for attribute value */
      entry->valptr = malloc(attrSize);

      if (entry->valptr) memset(entry->valptr, 0, attrSize);

      /* Store the attribute name */
#if defined(__CCI_SOLARIS__)
      CciCharCpy(entry->name, attrName);
#else
      wcscpy(entry->name, attrName);
#endif

      /* Link an existing previous element to this one */
      if (context->attrListPrevious) context->attrListPrevious->next = entry;
      else if ((context->attrListHead) == 0) context->attrListHead = entry;

      /* Save the pointer to the previous element */
      context->attrListPrevious = entry;
   }
   else {
         /* Error: Unable to allocate memory */
   }

   return(entry);

}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Utility function:                                              */
/*                                                                            */ 
/*  Acquire storage for an input terminal entry and link it to the context    */
/*                                                                            */
/******************************************************************************/
TERMINAL_LIST_ENTRY *insertInputTerminalListEntry(NODE_CONTEXT_ST *context, wchar_t *terminalName)
{
   TERMINAL_LIST_ENTRY *entry;
   int rc;

   entry = malloc(sizeof(TERMINAL_LIST_ENTRY));

   if (entry) {
      /* This entry is the current end of the list */
      entry->next = 0;

    /* Store the terminal name */
#if defined(__CCI_SOLARIS__)
      CciCharCpy(entry->name, terminalName);
#else
      wcscpy(entry->name, terminalName);
#endif

      /* Create terminal and save its handle */
      entry->handle = cniCreateInputTerminal(&rc, context->nodeObject, terminalName);

      /* Link an existing previous element to this one */
      if (context->inputTerminalListPrevious)
	      context->inputTerminalListPrevious->next = entry;
      else if ((context->inputTerminalListHead) == 0)
	      context->inputTerminalListHead = entry;

      /* Save the pointer to the previous element */
      context->inputTerminalListPrevious = entry;

   } /* if (entry) */

  return(entry);
}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Utility function:                                              */
/*                                                                            */ 
/*  Acquire storage for an output terminal and link it to the context         */
/*                                                                            */
/******************************************************************************/
TERMINAL_LIST_ENTRY *insertOutputTerminalListEntry(NODE_CONTEXT_ST *context, CciChar *terminalName)
{
   TERMINAL_LIST_ENTRY *entry;
   int rc;

   entry = malloc(sizeof(TERMINAL_LIST_ENTRY));
   if (entry) {
      /* This entry is the current end of the list */
      entry->next = 0;

      /* Store the terminal name */
#if defined(__CCI_SOLARIS__)
      CciCharCpy(entry->name, terminalName);
#else
      wcscpy(entry->name, terminalName);
#endif

      /* Create terminal and save its handle */
      entry->handle = cniCreateOutputTerminal(&rc, context->nodeObject, terminalName);

      /* Link an existing previous element to this one */
      if (context->outputTerminalListPrevious)
	      context->outputTerminalListPrevious->next = entry;
      else if ((context->outputTerminalListHead) == 0)
	      context->outputTerminalListHead = entry;

      /* Save the pointer to the previous element */
      context->outputTerminalListPrevious = entry;
   }

   return(entry);
}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Utility function:                                              */
/*                                                                            */ 
/*  Return the terminal object for a named input terminal                     */
/*                                                                            */
/******************************************************************************/
void *getInputTerminalHandle(NODE_CONTEXT_ST *context, CciChar *terminalName)
{
   static wchar_t* functionName=L"getInputTerminalHandle()";
   TERMINAL_LIST_ENTRY *p = ((NODE_CONTEXT_ST *)context)->inputTerminalListHead;

  /* Scan the list for a matching terminal name */
  while (p != 0) {
#if defined(__CCI_SOLARIS__)
      if (CciCharCmp(p->name, terminalName) == 0) {
#else
      if (wcscmp(p->name, terminalName) == 0) {
#endif
         /* Match found; return it */
         return(p->handle);
      }
      p = p->next;
   }

   /* No match was found */
   return(0);
}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Utility function:                                              */
/*                                                                            */ 
/*  Return the terminal object for a named output terminal                    */
/*                                                                            */
/******************************************************************************/
void *getOutputTerminalHandle(NODE_CONTEXT_ST *context, CciChar *terminalName)
{
   static wchar_t* functionName=L"getOutputTerminalHandle()";
   TERMINAL_LIST_ENTRY *p = ((NODE_CONTEXT_ST *)context)->outputTerminalListHead;

   /* Scan the list for a matching terminal name */
   while (p != 0) {
#if defined(__CCI_SOLARIS__)
      if (CciCharCmp(p->name, terminalName) == 0) {
#else
      if (wcscmp(p->name, terminalName) == 0) {
#endif
         /* Match found; return it */
         return(p->handle);
      }
      p = p->next;

   } /* while (p != 0) */

   /* No match was found */
   return(0);
}



/******************************************************************************/
/* Node Implementation Functions                                              */
/*                                                                            */
/* These are mandatory and MUST be implemented by the programmer.             */
/*                                                                            */
/* createNodeContext                                                          */
/* deleteNodeContext                                                          */
/* evaluate                                                                   */
/* getAttribute                                                               */
/* getAttriubteName                                                           */
/* setAttribute                                                               */
/* bipGetMessageflowNodeFactory                                               */
/*                                                                            */
/******************************************************************************/

/******************************************************************************/
/*                                                                            */
/* Plugin Node Implementation Function:                cniCreateNodeContext() */
/*                                                                            */
/******************************************************************************/
void *createNodeContext(CciFactory *factoryObject, CciChar *nodeName, CciNode *nodeObject)
{
   
   static wchar_t* functionName=L"createNodeContext()";
   NODE_CONTEXT_ST* p;


   /* Allocate a pointer to the local context */
   p = malloc(sizeof(NODE_CONTEXT_ST));

   /* If malloc was successful */
   if (p) {
      /* Clear the context area */
      memset(p, 0, sizeof(NODE_CONTEXT_ST)); /* Clear the context area */

      /* Save our node object pointer in our context */
	   p->nodeObject = nodeObject;

      /* Save our node name */
      wcsncpy((CciChar *)&p->nodeName, nodeName, MAX_NODE_NAME_LEN);

	   /* Verify the name of the node is supported by the factory */
	   if (wcscmp(nodeName, L"CheckDuplicationNode") == 0) {

		 insertAttributeTableEntry(p, nodeName, L"Root", CNI_TYPE_STRING);
		 insertAttributeTableEntry(p, nodeName, L"Database", CNI_TYPE_STRING);
		 insertAttributeTableEntry(p, nodeName, L"Table", CNI_TYPE_STRING);
		 insertAttributeTableEntry(p, nodeName, L"Field", CNI_TYPE_STRING);
		 insertAttributeTableEntry(p, nodeName, L"Correct", CNI_TYPE_STRING);

		 
		 /* Create terminals */
#if defined(__CCI_SOLARIS__)
         {
            const CciChar* ucsIn = CciString("in") ;
            insertInputTerminalListEntry(p, (CciChar*)ucsIn);
            free((void *)ucsIn) ;
         }
    
         {
            const CciChar* ucsOut = CciString("out") ;
            insertOutputTerminalListEntry(p, (CciChar*)ucsOut);
            free((void *)ucsOut) ;
         }

         {
            const CciChar* ucsFailure = CciString("failure") ;
            insertOutputTerminalListEntry(p, (CciChar*)ucsFailure);
            free((void *)ucsFailure) ;
         }
#else
	      insertInputTerminalListEntry(p, L"in");
	      insertOutputTerminalListEntry(p, L"out");
	      insertOutputTerminalListEntry(p, L"failure");
#endif
	   }
	   else{
	      /* Error handling if name of node is not supported by factory */
	   }
   }
   else {
      /* Error handling if malloc failed */
   }

   /* Return the pointer to the local context */
   return(p);
}


/******************************************************************************/
/*                                                                            */
/* Plugin Node Implementation Function:                cniDeleteNodeContext() */
/*                                                                            */
/* This function is called when an instance of a node is deleted.             */
/*                                                                            */
/******************************************************************************/
void deleteNodeContext(CciContext* context)
{
  static char* functionName = "deleteNodeContext()";

  return;
}


/******************************************************************************/
/*                                                                            */
/* Plugin Node Implementation Function:                     cniGetAttribute() */
/*                                                                            */
/******************************************************************************/
int getAttribute(CciContext *context, CciChar *attrName, CciChar *buffer, int bufsize)
{
   static wchar_t* functionName=L"getAttribute()";
   ATTRIBUTE_TBL_ENTRY* p = ((NODE_CONTEXT_ST *)context)->attrListHead;
   NODE_CONTEXT_ST*     nc = context;
   wchar_t              wstring[256];
   char                 string[256];  

   while (p != 0) {
      if (wcscmp(p->name, attrName) == 0) {
         switch(p->type) {
            case CNI_TYPE_INTEGER:
               sprintf((char*)&string, "%d", *((int*)p->valptr));
               mbstowcs((wchar_t*)&wstring, (char*)&string, 256); 
               wcsncpy(buffer, (wchar_t*)&wstring, min(bufsize, 256));
               break;
            case CNI_TYPE_STRING:
               memcpy(buffer, p->valptr, min(MAX_STRING_SIZE, bufsize));
            default:
            /* Error: Invalid datatype */
            break;
         }

         /* Normal return; attribute value returned */ 
         return(0);
      }

      p = p->next;
   }

   /* Attribute not owned by plugin */
   return(1); 
}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Implementation Function:                     cniSetAttribute() */
/*                                                                            */
/******************************************************************************/
int setAttribute(CciContext *context, CciChar *attrName, CciChar *attrValue)
{
   static wchar_t      *functionName = L"setAttribute()";
   ATTRIBUTE_TBL_ENTRY *p            = ((NODE_CONTEXT_ST *)context)->attrListHead;
   
   while (p != 0) {
#if defined(__CCI_SOLARIS__)
      if (CciCharCmp(p->name, attrName) == 0) {
#else
      if (wcscmp(p->name, attrName) == 0) {
#endif
         switch(p->type) {
            case CNI_TYPE_INTEGER:
#if defined(__CCI_SOLARIS__)
               {
                  const char* mbAttrValue = mbString(attrValue);
                  *((int *)p->valptr) = atoi(mbAttrValue);
                  free((void *)mbAttrValue);
               }
#elif defined(__CCI_AIX__)
               *((int *)p->valptr) = watoi(attrValue);
#else
               *((int *)p->valptr) = _wtoi(attrValue);
#endif
               break;
            case CNI_TYPE_STRING:
#if defined(__CCI_SOLARIS__)
               {
                  const char* mbAttrValue = mbString(attrValue);
                  memcpy(p->valptr, mbAttrValue, MAX_STRING_SIZE);
                  free((void *)mbAttrValue);
               }
#else
               memcpy(p->valptr, attrValue, MAX_STRING_SIZE); 
#endif
            default:
               /* Error: Invalid datatype */
               break;
         }

         /* Normal return; attribute value stored by plugin */
         return(0);
      }

      p = p->next;
   }

   /* Attribute not owned by plugin */
   return(1);
}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Implementation Function:                     cniGetAttribute() */
/*                                                                            */
/******************************************************************************/
int getAttributeName(CciContext* context, int index, CciChar* buffer, int bufferSize)
{
   static wchar_t      *functionName = L"getAttributeName()";
   int                  count        = 0;
   ATTRIBUTE_TBL_ENTRY *p            = ((NODE_CONTEXT_ST *)context)->attrListHead;
    
   while (p != 0) {
      if (count == index) {
#if defined(__CCI_SOLARIS__)
         CciCharCpy(buffer, p->name);
#else
         wcscpy(buffer, p->name);
#endif
         /* Normal return; attribute name returned */
         return(0);
      }
      count++;
      p = p->next;
   }

   /* Attribute not owned by the plugin node; defer to base */
   return(1);
}


/******************************************************************************/
/*                                                                            */
/* Plugin Node Implementation Function:                         cniEvaluate() */
/* This function provides to:                                                 */
/* - check the possible error of duplication in a database                    */
/* - correct it whenever it is possible                                       */
/* - track errors and copy a specific error message in the out message        */
/******************************************************************************/
void evaluate(CciContext *context, CciMessage *destinationList, CciMessage *exceptionList, CciMessage *message)
{

//init variables
	static wchar_t	*functionName = L"evaluate()";
   CCI_EXCEPTION_ST exception_st;

   int				error         = 1;
   int				rc            = 0;
   int				retvalue      = 1;
   int				EqualOffset   = 0;
   int				found;
   
   int				strLen;
   char				*ptr1 = NULL;
   char				*ptr2 = NULL;
   char				nextElementName[256];
   char				*root = NULL;
   char				*database = NULL;
   char				*table = NULL;
   char				*field = NULL;
   char				*state = NULL;
   char				CorrectChoice;
   char *val;
   CciChar *codemsg;	

   CciChar	*croot;
   CciChar	*db;
   CciChar	*statement;
   CciSqlExpression *sqlObject;

   CciValueType valueType;
   CciMessage		*data, *outMsg;
   CciTerminal		*outputTerminalObject = NULL;
   CciElement       *element;
   CciElement       *nextElement,*nextElement1;
   CciElement       *inRootElement, *outRootElement;
   CciElement		*rootElement;
   CciChar          elementName[256];
   CciChar			*value=NULL;
   CciChar			*valueChar[256];
   CciSize          elementNameSize=0;

   wchar_t          AttributeBuffer[256];
   wchar_t          *wNextElementName      = NULL;


/**********************************************************/
/* we get all the attribute value specified in the plugin */
/**********************************************************/

   // Get attribute value of "Root" attribute
   rc = getAttribute(context, L"Root", AttributeBuffer, sizeof(AttributeBuffer));
   root = UnicodeToASCII(AttributeBuffer);

   // Get attribute value of "Database" attribute
   rc = getAttribute(context, L"Database", AttributeBuffer, sizeof(AttributeBuffer));
   database = UnicodeToASCII(AttributeBuffer);

   // Get attribute value of "Table" attribute
   rc = getAttribute(context, L"Table", AttributeBuffer, sizeof(AttributeBuffer));
   table = UnicodeToASCII(AttributeBuffer);

   // Get attribute value of "Field" attribute
   rc = getAttribute(context, L"Field", AttributeBuffer, sizeof(AttributeBuffer));
   field = UnicodeToASCII(AttributeBuffer);

   // Get attribute value of "Correct" attribute
   rc = getAttribute(context, L"Correct", AttributeBuffer, sizeof(AttributeBuffer));
#if defined(__CCI_SOLARIS__)
   if (CciCharCmp(AttributeBuffer, L"Yes") == 0)
#else
   if (wcscmp(AttributeBuffer, L"Yes") == 0)
#endif
      CorrectChoice = CORRECTYES;
#if defined(__CCI_SOLARIS__)
   else if (CciCharCmp(AttributeBuffer, L"No") == 0)
#else
   else if (wcscmp(AttributeBuffer, L"No") == 0)
#endif
      CorrectChoice = CORRECTNO;


/**********************************************************************************/
/* Determine if the specified root exists:										  */
/* - Copy the original message in an out message								  */
/* - Get the root element of the message                                          */
/* - and compare with the specified root attribute in the plug-in                 */
/**********************************************************************************/

	// Get the root element
   inRootElement = cniRootElement(&rc, message);
   
	// Create a new message for output
   outMsg = cniCreateMessage(&rc, cniGetMessageContext(&rc, message));

	// Get the root element of the output message
   outRootElement = cniRootElement(&rc, outMsg);

	// Copy the contents of the input message to the output message
   cniCopyElementTree(&rc, inRootElement, outRootElement);


	// Get the last child of the root
   element = cniLastChild(&rc,outRootElement);

   found = 0;
   ptr1 = strstr(root, ".");

   // loop until the root is found
   while (ptr1) {
      ptr1++;
      ptr2 = strstr(ptr1, ".");
      if (ptr2)
         strLen = (int) (ptr2-ptr1);
      else
         strLen = strlen(root) - (int) (ptr1 - root);
      
	  memcpy(nextElementName, ptr1, strLen);
      nextElementName[strLen] = '\0';
      wNextElementName = ASCIIToUnicode(nextElementName);

      // Go to the first child of this element
      nextElement = cniFirstChild(&rc, element);
	  CCI_CHECK_RC();


      // Loop until element found or no element left
      found = 0;

      while (nextElement) {
      // Get the element's name
         elementNameSize = cniElementName(&rc, nextElement, (CciChar*)&elementName, sizeof(elementName));

	  // Check if name is right
         if ( wcscmp(elementName, wNextElementName) == 0 ) {
            found = 1;
            break;
         }
         nextElement = cniNextSibling(&rc, nextElement);
      }
      if (found == 1) {
      // Element found: go to this child
         element = nextElement;
      }
      else {

		  found = 0;
         break;
      }
      ptr1 = ptr2;
   }
   //end of the loop

   outputFile   = fopen("c:\\CheckDuplicationDebug.txt", "w");
   fprintf(outputFile,"found%d\n",found);
   fflush(outputFile);
   // if the root path exists: read root value from this element 
   if (found == 1) {
      elementNameSize=cniElementCharacterValue(&rc, element, (CciChar*)&valueChar, sizeof(valueChar));
   // convert from char* to CciChar	
	  db= CciString(database);
   // init statement memory
	  state = (char*)malloc(sizeof(char) * (150));

   // if the root id has a specified value  
	  if(elementNameSize > 1){
         root = UnicodeToASCII(valueChar);
		 fprintf(outputFile,"root %s\n",root);
		 fflush(outputFile);

		 val=strstr(root,".");

   // it is possible to have important problems if integer values have got spaces
   // when we try to select value.
   // I suggest either to clear spaces here or not to use spaces
	     if(elementNameSize > 11)
		 {
			fprintf(outputFile,"first test size 1\n");
			fprintf(outputFile,"the value inserted is to big for the database (more than 2147483647) %S",valueChar);
			fflush(outputFile);
			error=2;

   // create XML tag and insert message error
			codemsg=L"CheckDuplicationError";
			value=L"The maximum we can insert in the database with integer primary key is 2147483647.";
			outMsg = CheckDuplicationMsgError(outMsg, value, codemsg);

		 }
		 else if( strcmp(root,"2147483647") > 0 && elementNameSize == 11){
			fprintf(outputFile,"first test size 2\n");
			fprintf(outputFile,"the value inserted is to big for the database (more than 2147483647) %S",valueChar);
			fflush(outputFile);
			error=2;

   // create XML tag and insert message error
			codemsg=L"CheckDuplicationError";
			value=L"The maximum we can insert in the database with integer primary key is 2147483647.";
			outMsg = CheckDuplicationMsgError(outMsg, value, codemsg);
		 }
		 else if (val = strstr(root,".") != NULL){
			value=L"Bad data type of the id message (not integer)";
			codemsg=L"CheckDuplicationError";
		    outMsg = CheckDuplicationMsgError(outMsg, value, codemsg);
			error = 2;
		 }
		 else{
			 
   // create the select statement and check if value already exists in the database
	     state=strcpy(state,"SET OutputRoot.XML = THE(SELECT A.");
	     state=strcat(state,field);
	     state=strcat(state," FROM Database.");
	     state=strcat(state,table);
	     state=strcat(state," AS A WHERE A.");
	     state=strcat(state,field);
	     state=strcat(state,"=");
	     state=strcat(state,root);
	     state=strcat(state,");");
   // convert from char* to CciChar	
         statement = CciString(state);
		 
		 fprintf(outputFile,"statement= %S",statement);
		 fflush(outputFile);

   //create sqlexpression	  
		 sqlObject = cniSqlCreateStatement(&rc,((NODE_CONTEXT_ST *)context)->nodeObject,db,CCI_SQL_TRANSACTION_AUTO,statement);
   if (rc != CCI_SUCCESS)
	{
  	  error=2;
	  fprintf(outputFile,"cniSqlCreateStatement failed with rc = %d\n", rc);
//function call failed
        if ( rc >= CCI_EXCEPTION )
        {            
            memset(&exception_st, 0, sizeof(exception_st));
            cciGetLastExceptionData(&rc, &exception_st);
			if (exception_st.type == 8)
				value=L"Database error: the database name does not exist.";
			else
				value=L"Database error: it can be a wrong table or field name, or a bad syntax of the id value (spaces,character)";
			codemsg=L"CheckDuplicationError";
		    outMsg = CheckDuplicationMsgError(outMsg, value, codemsg);
        }
		fflush(outputFile);
	}

   if(error != 2){	
//Create the message saving the data returned by the request
//Execute the sql "select" request
		  data = cniCreateMessage(&rc,cniGetMessageContext(&rc,message));
          cniSqlSelect(&rc,sqlObject,destinationList,exceptionList,message,data);
   if (rc != CCI_SUCCESS)
	{

	  error=2;
	  fprintf(outputFile,"cniSqlselect failed with rc = %d\n", rc);

//function call failed
        if ( rc >= CCI_EXCEPTION )
        {            
            memset(&exception_st, 0, sizeof(exception_st));
            cciGetLastExceptionData(&rc, &exception_st);
			if (exception_st.type == 8)
				value=L"Database error: the database name does not exist.";
			else
				value=L"Database error: it can be a wrong table or field name, or a bad syntax of the id value (spaces,character)";
			codemsg=L"CheckDuplicationError";
		    outMsg = CheckDuplicationMsgError(outMsg, value, codemsg);
        }
		fflush(outputFile);
	}
	}
//Delete the object
          cniSqlDeleteStatement(&rc,sqlObject);

		  if (error != 2){

// go through the message data to get the data
		  rootElement = cniRootElement(&rc, data);
		  nextElement1 = cniLastChild(&rc, rootElement);
		  nextElement1 = cniFirstChild(&rc, nextElement1);

   //Get the type of the element value
		  valueType = cniElementValueType(&rc, nextElement1);

   // test the value of the type
		  switch (valueType){
			  case CCI_VALUE_TYPE_CHAR:
              fprintf(outputFile,"element value type is CHAR\n");
			  fprintf(outputFile,"1 type eq 3\n");
   // if the value found in the database is the same than the value in the message
			  fprintf(outputFile,"2 type eq 3\n");
		      elementNameSize = cniElementCharacterValue(&rc,nextElement1,(CciChar*)&valueChar,sizeof(valueChar));
			  fprintf(outputFile, "first select value = %S\nLength = %d\n", valueChar,elementNameSize);
			  fflush(outputFile);
			  error = 0;

			  croot=CciString(root);
			  if (CciCharCmp(valueChar, croot) == 0){
				  error = 1;
				  fprintf(outputFile,"duplication found\n value of correct %d\n",CorrectChoice);
				  fflush(outputFile);
	// if no correction is required
				  if (CorrectChoice != CORRECTYES){
   //CheckDuplication error : Duplication of the primary key in the database
					  codemsg=L"CheckDuplicationError";
					  value=L"Duplication error. And no automatic correction required";
					  outMsg = CheckDuplicationMsgError(outMsg, value, codemsg);
				  }
			  }
			  break;

   // if the database is empty			  
			  case 0:
			  fprintf(outputFile,"database is empty\n");
			  error=0;
			  break;

   // If the type does not correspond to the data tested in the message, there will be a error
			  default:
              fprintf(outputFile,"element value type is (default) UNKNOWN or no relevant format\n");
   //CheckDuplication error: Wrong type in the database
			  error = 2;
			  codemsg=L"CheckDuplicationError";
		      value=L"Wrong data type in the database.";
		      outMsg = CheckDuplicationMsgError(outMsg, value, codemsg);
			  break;
		  }

		  }
		}
	  }
   // if no value specified in the message
	  else{
		fprintf(outputFile,"\nno value specifed in msg");
		error=1;
		CorrectChoice=CORRECTYES;
	  }
	}
   // if the specified root in the plugin does not exist in the message
	else{
	    error=2;
   // error CheckDuplication: the specified root id does not exist
		codemsg=L"CheckDuplicationError";
		value=L"the specified root id does not exist in the message.";
		outMsg = CheckDuplicationMsgError(outMsg, value, codemsg);

	
	}

   // if Correct is checked and the error is 1.
   if (error == 1 && CorrectChoice == CORRECTYES){

   // create the select statement to take the max value in the database and add 1
		state=strcpy(state,"SET OutputRoot.XML.");
		state=strcat(state,field);
		state=strcat(state,"= THE(SELECT MAX(A.");
		state=strcat(state,field);
		state=strcat(state,") FROM Database.");
		state=strcat(state,table);
		state=strcat(state," AS A) +1;");

   // convert from char* to CciChar	
        statement = CciString(state);

		fprintf(outputFile,"statement 2= %S",statement);
		fflush(outputFile);


   //create sqlexpression
        sqlObject = cniSqlCreateStatement(&rc,((NODE_CONTEXT_ST *)context)->nodeObject,db,CCI_SQL_TRANSACTION_AUTO,statement);
   if (rc != CCI_SUCCESS)
	{
	  error=2;
	  fprintf(outputFile,"cniSqlCreateStatement failed with rc = %d\n", rc);

//function call failed
        if ( rc >= CCI_EXCEPTION )
        {            
            memset(&exception_st, 0, sizeof(exception_st));
            cciGetLastExceptionData(&rc, &exception_st);
			if (exception_st.type == 8)
				value=L"Database error: the database name does not exist.";
			else
				value=L"Database error: it can be a wrong table or field name, or a bad syntax of the id value (spaces,character)";
			codemsg=L"CheckDuplicationError";
		    outMsg = CheckDuplicationMsgError(outMsg, value, codemsg);
        }
		fflush(outputFile);
	}

   //Create the message saving the data returned by the request
   //Execute the sql "select" request
        data = cniCreateMessage(&rc,cniGetMessageContext(&rc,message));
        cniSqlSelect(&rc,sqlObject,destinationList,exceptionList,message,data);
   if (rc != CCI_SUCCESS)
	{
	  error=2;
	  fprintf(outputFile,"cniSqlSelect failed with rc = %d\n", rc);

//function call failed
        if ( rc >= CCI_EXCEPTION )
        {            
            memset(&exception_st, 0, sizeof(exception_st));
            cciGetLastExceptionData(&rc, &exception_st);
			if (exception_st.type == 8)
				value=L"Database error: the database name does not exist.";
			else
				value=L"Database error: it can be a wrong table or field name, or a bad syntax of the id value (spaces,character)";
			codemsg=L"CheckDuplicationError";
		    outMsg = CheckDuplicationMsgError(outMsg, value, codemsg);
        }
		fflush(outputFile);
	}

   //Delete the object
        cniSqlDeleteStatement(&rc,sqlObject);

   // go through the message data to get the data
	    rootElement = cniRootElement(&rc, data);
        nextElement1 = cniLastChild(&rc, rootElement);
        nextElement1 = cniFirstChild(&rc, nextElement1);

   //Get the type of the element value
        valueType = cniElementValueType(&rc, nextElement1);
   
   // test the value of the type
		switch (valueType){
			  case CCI_VALUE_TYPE_CHAR:
	          fprintf(outputFile,"element value type is CHAR\n");

			  elementNameSize = cniElementCharacterValue(&rc,nextElement1,(CciChar*)&valueChar,sizeof(valueChar));
    		  fprintf(outputFile, "second select value = %S\nLength = %d", valueChar,elementNameSize);
    		  fflush(outputFile);
              root = UnicodeToASCII(valueChar);
	          if(elementNameSize > 11){
					fprintf(outputFile,"first test size 1\n");
					fprintf(outputFile,"the value inserted is to big for the database (more than 2147483647) %S",valueChar);
					fflush(outputFile);
					error=2;

   // create XML tag and insert message error
					codemsg=L"CheckDuplicationError";
					value=L"The maximum we can insert in the database with integer primary key is 2147483647.";
					outMsg = CheckDuplicationMsgError(outMsg, value, codemsg);

			  }
			  else if( strcmp(root,"2147483647") > 0 && elementNameSize == 11){
					fprintf(outputFile,"first test size 2\n");
					fprintf(outputFile,"the value inserted is to big for the database (more than 2147483647) %S",valueChar);
					fflush(outputFile);
					error=2;

   // create XML tag and insert message error
					codemsg=L"CheckDuplicationError";
					value=L"The maximum we can insert in the database with integer primary key is 2147483647.";
					outMsg = CheckDuplicationMsgError(outMsg, value, codemsg);
			  }
			  else{
					cniSetElementCharacterValue(&rc,element,valueChar,elementNameSize - 1);
					CCI_CHECK_RC();		
					fprintf(outputFile,"Copy max + 1\n");
					error = 0;
			  }
			  break;

   // if the database is empty			  
			  case 0:
			  cniSetElementCharacterValue(&rc,element,L"1",1);
			  CCI_CHECK_RC();
			  fprintf(outputFile,"database empty\n");
			  error = 0;
			  break;

   // if other type
			  default:
				  fprintf(outputFile,"element value type is (default) UNKNOWN or no relevant format %d,\n",valueType);
					fflush(outputFile);
   //CheckDuplication error: Wrong type in the database
				  error = 2;
		   		  codemsg=L"CheckDuplicationError";
			      value=L"Wrong data type in the database.";
				  outMsg = CheckDuplicationMsgError(outMsg, value, codemsg);
				  break;
		 }

   // the message to send is the outMsg
fclose(outputFile);
   }
message=outMsg;

   /* Get handle of output terminal */
   /* If there is no error */
	
if (error == 0)
{   
#if defined(__CCI_SOLARIS__)
      outputTerminalObject = getOutputTerminalHandle(context, (CciChar*)constOut);
#else
      outputTerminalObject = getOutputTerminalHandle(context, L"out");   
#endif
}
else
{
#if defined(__CCI_SOLARIS__)
      outputTerminalObject = getOutputTerminalHandle(context, (CciChar*)constFailure);
#else
      outputTerminalObject = getOutputTerminalHandle(context, L"failure");
#endif
}
   /* If the terminal exists and is attached, propagate to it */
   if (outputTerminalObject) {
      if (cniIsTerminalAttached(&rc, outputTerminalObject)) {
         if (rc == CCI_SUCCESS) {
            retvalue = cniPropagate(&rc, outputTerminalObject, destinationList, exceptionList, message);
            if (rc != CCI_SUCCESS) {
               if (rc == CCI_EXCEPTION) {
                  /* Get details of the exception */
                  memset(&exception_st, 0, sizeof(exception_st));
                  cciGetLastExceptionData(&rc, &exception_st);

                  /* Any local error handling may go here */

                  /* If the exception is "caught" it must be re-thrown */
                  cciRethrowLastException(0);
               }

               /* Some other error...log it */
               cciLog(&rc, CCI_LOG_ERROR, __FILE__, __LINE__, "",
#if defined(__CCI_SOLARIS__)
                 (CciChar*)constMQSeriesIntegrator2, 0, "Propagate failed", 0);
#else
                 L"MQSIv201", 0, "Propagate failed", 0);
#endif
            }
            else {
               /* Message was propagated successfully */
               cciLog(&rc, CCI_LOG_INFORMATION, __FILE__, __LINE__, "",
#if defined(__CCI_SOLARIS__)
                 (CciChar*)constMQSeriesIntegrator2, 0, "Message propagated", 0);
#else 
                 L"MQSIv201", 0, "Message propagated", 0);
#endif
            }
         }
         else {
            if (rc == CCI_EXCEPTION) {
               cciRethrowLastException(0);
            }
            else {
               /* Some other error...log it */
               cciLog(&rc, CCI_LOG_ERROR, __FILE__, __LINE__, "",
#if defined(__CCI_SOLARIS__)
                 (CciChar*)constMQSeriesIntegrator2, 0, "Error occurred", 0);
#else 
                 L"MQSIv201", 0, "Error occurred", 0);
#endif 
            }
         }
      }
   }
   else {
      /* Terminal did not exist...severe internal error */
      cciLog(&rc, CCI_LOG_ERROR, __FILE__, __LINE__, "",
#if defined(__CCI_SOLARIS__)
           (CciChar*)constMQSeriesIntegrator2,0, "Terminal not connected", 0);
#else 
           L"MQSIv201", 0, "Terminal not connected", 0);
#endif 
   }

   return;
}

#if defined(__CCI_SOLARIS__)
/******************************************************************************/
/*                                                                            */
/* Plugin Node Utility function:                                              */
/*                                                                            */
/*  initNodeConstants                                                         */
/*                                                                            */
/*  Creates Cci constant strings from defined char * constant strings.        */
/*                                                                            */
/******************************************************************************/
void initNodeConstants(
  void
){
  //constZero = CciString(CONST_ZERO);
  constMQSeriesIntegrator2 = CciString(CONST_MQSERIES_INTEGRATOR_2);
  constPlugInNode = CciString(CONST_PLUGIN_NODE);
  constPluginNodeFactory = CciString(CONST_PLUGIN_NODE_FACTORY);
  constOut = CciString(CONST_OUT);
  constFailure = CciString(CONST_FAILURE);
}
#endif


/******************************************************************************/
/*                                                                            */
/* Plugin Node Implementation Function:                                       */
/*                                                                            */
/* This function is called when the message broker loads the "lil" during the */
/* initialization of the broker engine.                                       */
/*                                                                            */
/* The responsibilities of the plug-in are to:                                */
/*  - create the node factory (or factories) which the "lil" supports, using  */
/*    the createNodeFactory() function, saving the returned pointer to the    */
/*    node factory for use at runtime.                                        */
/*  - define each node supported by the (or each) factory) and pass a pointer */
/*    to a function table which contains pointers to the node implementation  */
/*    functions.                                                              */
/*  - define the name of each attribute supported by each node, including a   */
/*    pointer to a validation function.                                       */
/*  - define input and output terminals supported by each node.               */
/*  - return the pointer to the factory, unless an error occurred, in which   */
/*    case zero should be returned.                                           */    
/*                                                                            */
/******************************************************************************/
CciFactory LilFactoryExportPrefix * LilFactoryExportSuffix *bipGetMessageflowNodeFactory()
{
   static wchar_t*   functionName = L"bipGetMessageflowNodeFactory()";
   CciFactory*       factoryObject;
   int               rc = 0;
   CciChar           factoryName[] = L"CheckDuplicationNodeFactory";
   CCI_EXCEPTION_ST  exception_st;
   static CNI_VFT    vftable;

#if defined(__CCI_SOLARIS__)
   /* Before we proceed we need to initialise all the static constants */
   /* that may be used by the plug-in.                                 */
   initNodeConstants();
#endif
   /* Create the Node Factory for this plug-in */
#if defined(__CCI_SOLARIS__)
   factoryObject = cniCreateNodeFactory(0, (unsigned short *)constPluginNodeFactory);
#else
   factoryObject = cniCreateNodeFactory(0, factoryName);
#endif

   if (factoryObject == CCI_NULL_ADDR) {
      if (rc == CCI_EXCEPTION) {
         /* Get details of the exception */
         memset(&exception_st, 0, sizeof(exception_st));
         cciGetLastExceptionData(&rc, &exception_st);

         /* Any local error handling may go here */

         /* Rethrow the exception */
         cciRethrowLastException(&rc);
      }
      /* Any further local error handling can go here */
   }
   else {
      /* Setup function table with pointers to node implementation functions */
      memset(&vftable, 0, sizeof(vftable));
      vftable.iFpCreateNodeContext     = createNodeContext;
      vftable.iFpDeleteNodeContext     = deleteNodeContext;
      vftable.iFpGetAttributeName      = getAttributeName;
      vftable.iFpSetAttribute          = setAttribute;
      vftable.iFpGetAttribute          = getAttribute;
      vftable.iFpEvaluate              = evaluate;

      /* Define a node type supported by our factory. If any errors/exceptions    */
      /* occur during the execution of this utility function, then as we have not */
      /* supplied the returnCode argument, the exception will bypass the plugin   */
      /* and be directly handled by the broker.                                   */
#if defined(__CCI_SOLARIS__)
      cniDefineNodeClass(0, factoryObject, (CciChar*)constPlugInNode, &vftable);
#else
      cniDefineNodeClass(0, factoryObject, L"CheckDuplicationNode", &vftable); 
#endif

   }

   /* Return address of this factory object to the broker */
   return(factoryObject);
}

char* UnicodeToASCII(const wchar_t* lStr)
{
   unsigned int len = wcslen(lStr);
   char* outStr = (char*)malloc(sizeof(char) * (len + 1));

   wcstombs(outStr, lStr, len + 1);
   outStr[len] = '\0';

  return outStr;
}


wchar_t* ASCIIToUnicode(const char* aStr)
{
   unsigned int len = strlen(aStr);
   wchar_t *outStr = (wchar_t*)malloc(sizeof(wchar_t) * (len + 1));
   mbstowcs(outStr, aStr, len + 1);
   outStr[len] = L'\0';
   return outStr;
}


wchar_t* ASCIInToUnicode(const char* aStr, size_t len)
{
   wchar_t *outStr = (wchar_t*)malloc(sizeof(wchar_t) * (len + 1));
   mbstowcs(outStr, aStr, len + 1);
   outStr[len] = L'\0';
   return outStr;
}


void wgetNameValue(char *Env, CciChar **Name, CciChar **Value)
{
   int i = 0;
   while ( Env[i] != '\0' && Env[i] != '=' ) i++;
      
   *Name  = ASCIInToUnicode(Env, i);
   *Value = ASCIIToUnicode(&(Env[i + 1])); 
}


CciChar *wStrUpper(CciChar *string)
{
   int i = 0;
   while ( ( string[i] = (char)towupper(string[i]) ) != L'\0' )
      i++;
   return(string);
}


CciChar *wStrLower(CciChar *string)
{
   int i = 0;
   while ( ( string[i] = (char)tolower(string[i]) ) != L'\0' )
      i++;
   return(string);
}

CciMessage *CheckDuplicationMsgError(CciMessage *message, CciChar *msg, CciChar *codemsg){
	int rc;
	CciElement *element;

	element=cniRootElement(&rc, message);
	element = cniLastChild(&rc,element);
	element=cniCreateElementAsLastChild(&rc,element);

	cniSetElementType(&rc, element, CCI_ELEMENT_TYPE_NAME);

	cniSetElementName(&rc, element, codemsg);

	cniSetElementCharacterValue(&rc, element, (const CciChar *)msg,wcslen(msg));
	
	return message;

}
